@@ -5,20 +5,22 @@
 # the BSD License: http://www.opensource.org/licenses/bsd-license.php
 # flake8: noqa
 #@PydevCodeAnalysisIgnore
+from git.exc import *                       # @NoMove @IgnorePep8
 import inspect
 import os
 import sys
-
 import os.path as osp
 
+from typing import Optional
+from git.types import PathLike
 
-__version__ = '3.1.13'
+__version__ = '3.1.17'
 
 
 #{ Initialization
-def _init_externals():
+def _init_externals() -> None:
     """Initialize external projects by putting them into the path"""
-    if __version__ == '3.1.13' and 'PYOXIDIZER' not in os.environ:
+    if __version__ == '3.1.17' and 'PYOXIDIZER' not in os.environ:
         sys.path.insert(1, osp.join(osp.dirname(__file__), 'ext', 'gitdb'))
 
     try:
@@ -29,13 +31,13 @@ def _init_externals():
 
 #} END initialization
 
+
 #################
 _init_externals()
 #################
 
 #{ Imports
 
-from git.exc import *                       # @NoMove @IgnorePep8
 try:
     from git.config import GitConfigParser  # @NoMove @IgnorePep8
     from git.objects import *               # @NoMove @IgnorePep8
@@ -65,7 +67,8 @@ __all__ = [name for name, obj in locals().items()
 #{ Initialize git executable path
 GIT_OK = None
 
-def refresh(path=None):
+
+def refresh(path: Optional[PathLike] = None) -> None:
     """Convenience method for setting the git executable path."""
     global GIT_OK
     GIT_OK = False
@@ -78,6 +81,7 @@ def refresh(path=None):
     GIT_OK = True
 #} END initialize git executable path
 
+
 #################
 try:
     refresh()
